# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pywordle']

package_data = \
{'': ['*'], 'pywordle': ['words/*']}

install_requires = \
['marshmallow', 'pytz', 'requests']

entry_points = \
{'console_scripts': ['wordle = pywordle.wordle:main']}

setup_kwargs = {
    'name': 'pywordle',
    'version': '0.3.1',
    'description': 'Little helper to play the Wordle game',
    'long_description': '# PyWordle\n\nThis package can help you playing the [Wordle](https://www.powerlanguage.co.uk/wordle/) game. It will not play for you. You will still have to choose the best strategy.\n\n## Installation\n\nThe tool can easily be installed on a computer where a recent (>=3.10) version of Python is present.\n\n```bash\npython3 -m venv venv\n. venv/bin/activate\npip install pywordle\n```\n\n## Usage\n\nEach time that you want to enter a new word, you will ask the tool to provide a list of candidates. When you have selected a word, type it in as well as the result provided by the `Wordle` game. The code values for the result colours are:\n\n* Dark gray: 0\n* Yellow: 1\n* Green: 2\n\nHere is an example:\n\n```bash\n$ wordle --letters=etaions --unique\nHere is the list of 37 possibilities: {\'aeons\': 2.72, \'onset\': 2.87, \'saint\': 4.43, \'tinea\': 0, \'oaten\': 0, \'antes\': 3.17, \'siena\': 2.4, \'atone\': 3.11, \'tines\': 2.7, \'steno\': 1.97, \'stoae\': 0, \'anise\': 2.3, \'taino\': 0, \'stain\': 3.78, \'stoai\': 0, \'notes\': 4.36, \'sotie\': 0, \'iotas\': 2.68, \'tenia\': 2.73, \'neats\': 3.93, \'nates\': 0, \'satin\': 3.12, \'etnas\': 2.43, \'senti\': 1.97, \'antis\': 4.33, \'stone\': 4.68, \'seton\': 1.73, \'tsine\': 0, \'entia\': 0, \'eosin\': 1.74, \'ostia\': 1.82, \'tones\': 4.27, \'tains\': 2.2, \'inset\': 1.73, \'noise\': 4.6, \'stane\': 0, \'stein\': 3.38}\nPlease enter the word played:\nnoise\nPlease enter the result:\n20001\nSaving result for \'noise\': [2, 0, 0, 0, 1]\n```\n\nThe result will be stored in today\'s game file and be used next time you call the tool, limiting the possibilities. Just type `Enter`, if you do not want to choose a word just yet. Each day a file `games/YYYYMMDD.json` will be created to hold your progress.\n\nAll the command line options can be found with the `--help` option.\n\n```bash\n$ wordle --help\nusage: Propose words for the Wordle game. [-h] [--large] [--stats] [--none] [--unique] [--check] [--verbose] [--letters LETTERS] [--minfrequency MINFREQUENCY]\n\noptions:\n  -h, --help            show this help message and exit\n  --large               Uses the large file of English words.\n  --stats               Displays some stats about the found words.\n  --none                Use none of the previously used letters.\n  --unique              Do not repeat letters in a word.\n  --check               Check that the word is valid.\n  --verbose             Print progress.\n  --sort                Make sure the words are in alphabetic order.\n  --letters LETTERS     The set of letters to be used.\n  --minfrequency MINFREQUENCY\n                        A minimum frequency for the proposed words, either 0 (not found) or between 1 and 7.\n\n```\n\n### Letters\n\nThe `--letters` is useful to start the game with a word that is using the most common English letters to have the best chances to find letters of the solution word. It also could be a good idea to add the `--unique` option to avoid repeating a letter.\n\n```bash\nwordle --letters=etaionshr --unique\n```\n\n### None\n\nThis is a different strategy. Instead of looking for a possible word, you are trying to look for information on letters that you have not used yet today. This is not possible if you have selected the `Hard Mode`. It only makes sense to use the option for the second and possibly third word. Again the `--unique` option will give you more information.\n\n```bash\nwordle --none --unique\n```\n\n### Stats\n\nThis will give you some statistics in the list of proposed words. First you will get the most used letters in that list. Then you will get the words that used the most used letters. This is another strategy to give you more chance to have valuable information.\n\n```bash\n$ wordle --letters=etaion --unique --stats\nHere is the list of 6 possibilities: {\'tinea\': 0, \'oaten\': 0, \'atone\': 3.11, \'taino\': 0, \'tenia\': 2.73, \'entia\': 0}\n10 most used letters in the words: a, n, t, e, i, o, b, c, d, f\n5 most scored words: tinea, tenia, entia, oaten, atone\nPlease enter the word played:\n\nNo play recorded.\n```\n\n### Using the Words API\n\nIn order to check the validity and frequency of the words with a call to the [Words API](https://github.com/dwyl/english-words), you need to store a valid [Rapid API](https://rapidapi.com) key in `RAPIDAPI_KEY` environment variable. It is free for up to 2500 words per day. Once a word is checked, it is stored locally so it will not be checked again the next days. When there is no more "null" values in your JSON word file, you do not need to check the Words API anymore.\n\n```bash\nexport RAPIDAPI_KEY="{YourKey}"\nwordle --check\n```\n\n### Frequency\n\nWhen checking a word in the API, we store its frequency. The solution of the Wordle game tends to be a word with a high frequency.\n\n```bash\nwordle --minfrequency=5\n```\n\n### Large English word list\n\nBy default the tool will rely on a list that contains less than 4300 words of 5 letters. The good news is that most are valid but a few are missing. With the `--large` option, you will use a list with nearly 16000 words of 5 letters. It\'s exhaustive but unfortunately it also contains many invalid words. This is where the Words API will become handy. But beware of your daily quota.\n\nIt is critical to choose at the first run of the tool if you decide to start from the large list or not. After that, the option will be ignored and we will only use the generated JSON file: `words/words.en.json`. If you delete it to start again, you will lose the benefit of all the checks you have done with the Words API.\n\n```bash\nwordle --large\n```\n',
    'author': 'Pierre Cart-Grandjean',
    'author_email': 'pcart-grandjea@noreply.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
"""
Play Wordle game
https://www.powerlanguage.co.uk/wordle/

Usage:
poetry install
poetry run wordle --help
"""

from argparse import ArgumentParser, Namespace
from collections import OrderedDict
from datetime import datetime
from importlib.resources import open_text
from json import dump, loads
from os import getenv, makedirs
from os.path import join
from string import ascii_letters
import sys
from typing import Any, Dict, List, Generator, Callable, Union, Optional

from marshmallow import Schema, fields, post_load
from pytz import timezone
from requests import get

WORDS_DIR = "words"
WORDS_FOUND_FILE = "words.en.json"
WORDS_NOT_FOUND_FILE = "words.en.notfound.txt"
GAME_FILE = f"games/{datetime.now(timezone('CET')).strftime('%Y%m%d')}.json"
ARGS = Namespace()
WORDS_API: Dict[str, int] = {"counter": 0, "quota": 2500}


# pylint: disable=no-member, no-self-use


def parse_arguments() -> None:
    """
    Parse the command line arguments.
    """
    parser = ArgumentParser("Propose words for the Wordle game.")
    parser.add_argument("--large", action="store_true", help="Uses the large file of english words.")
    parser.add_argument("--stats", action="store_true", help="Displays some stats about the found words.")
    parser.add_argument("--none", action="store_true", help="Use none of the previously used letters.")
    parser.add_argument("--unique", action="store_true", help="Do not repeat letters in a word.")
    parser.add_argument("--check", action="store_true", help="Check that the word is valid.")
    parser.add_argument("--verbose", action="store_true", help="Print progress.")
    parser.add_argument("--sort", action="store_true", help="Make sure the words are in alphabetic order.")
    parser.add_argument("--letters", help="The set of letters to be used.")
    parser.add_argument(
        "--minfrequency",
        type=float,
        default=1,
        help="A minimum frequency for the proposed words, either 0 (not found) or between 1 and 7.",
    )
    parser.parse_args(namespace=ARGS)
    if ARGS.check and getenv("RAPIDAPI_KEY") is None:
        print(
            "You need to provide a valid Rapid API key in RAPIDAPI_KEY environment variable",
            "to be able to check the existence and the usage frequency of the words.",
        )
        sys.exit(1)


class Play:
    """
    This represent a played word with its result.
    For each letter is the corresponding code:
    0: letter not in word
    1: letter in word in a different position
    2: letter in word in that position
    """

    word: str
    result: List[int]

    def __init__(self, word: str, result: List[int]):
        self.word = word
        self.result = result

    def is_valid(self, word: str) -> bool:  # NOSONAR
        """
        Check if the new word would give this result for the current word.
        """
        if len(self.word) != len(word):
            return False
        letters = list(self.word)
        newletters = list(word)
        # First handle the exact match
        for index, letter in enumerate(letters):
            if self.result[index] == 2:
                if newletters[index] != letter:  # Exact match is missing
                    return False
                letters[index] = "_"
                newletters[index] = "_"
        # Now handle the non exact match
        for index, letter in enumerate(letters):
            if self.result[index] == 1:
                if newletters[index] == letter:  # This is an exact match where it should not
                    return False
                if letter not in newletters:  # We should find this letter in the solution
                    return False
                letters[index] = "_"
                newindex = newletters.index(letter)
                newletters[newindex] = "_"
        # Finally handle the miss
        for index, letter in enumerate(letters):
            if self.result[index] == 0 and letter in newletters:  # There should be no match
                return False
        # It's all good, we have a match
        return True

    def is_none(self, word: str) -> bool:
        """
        Check if the proposed word has none of the letters of the played word.
        """
        for letter in word:
            # Letter is never used before
            if letter in self.word:
                return False
        return True


class Game:
    """
    This is a full game session with the list of words played already.
    """

    played: List[Play]

    def __init__(self, played: List[Play]):
        self.played = played

    def is_valid(self, word: str) -> bool:
        """
        Check each word played to see if the provided one is a match.
        """
        for play in self.played:
            if not play.is_valid(word):
                return False
        return True

    def is_none(self, word: str) -> bool:
        """
        Check each word played to see if the provided one uses any of the letters.
        """
        for play in self.played:
            if not play.is_none(word):
                return False
        return True


class PlaySchema(Schema):
    """
    Marshmallow schema of the Play object.
    """

    word = fields.Str()
    result = fields.List(fields.Integer)

    @post_load
    def make_play(self, data: Any, **_: Any) -> Play:
        """
        This mill create a Play object when invoking a load() on the schema.
        """
        return Play(**data)


class GameSchema(Schema):
    """
    Marshmallow schema of the Game object.
    """

    played = fields.List(fields.Nested(PlaySchema))

    @post_load
    def make_game(self, data: Any, **_: Any) -> Game:
        """
        This will create a Game object when invoking a load() on the schema.
        """
        return Game(**data)


def get_game(gamefilename: str) -> Union[Game, Any]:
    """
    Get the current game progress from a JSON file.
    """
    try:
        with open(gamefilename, mode="r", encoding="utf_8") as gfile:
            return GameSchema().loads(gfile.read())
    except FileNotFoundError:
        # If the game file was not found, we create it
        # First we make sure that the games directory exists
        makedirs("games", exist_ok=True)
        emptygame: dict[str, List[Any]] = {"played": []}
        with open(gamefilename, mode="w", encoding="utf_8") as gfile:
            dump(emptygame, gfile, indent=4)
        return GameSchema().load(emptygame)


def get_words_txt() -> Generator[str, None, None]:
    """
    Read the words from the words.en.txt file.
    It is now a ressource provided by the package.
    """
    wordsfilename = f"words.{'large' if ARGS.large else 'small'}.en.txt"
    with open_text("pywordle.words", wordsfilename) as wfile:
        for line in wfile:
            yield line.split("\n")[0]


def save_words(words: Dict[str, Optional[int]]) -> None:
    """
    Save the words
    """
    # Make sure the words directory exists
    makedirs(WORDS_DIR, exist_ok=True)
    with open(join(WORDS_DIR, WORDS_FOUND_FILE), mode="w", encoding="utf_8") as jsonfile:
        dump(words, jsonfile, indent=4)


def get_words() -> Dict[str, Optional[int]]:
    """
    Read the words from the local JSON file that already has the already fetched
    frequency from Words API.
    If empty, create one from the text file.
    """
    # Get the list of words from the local JSON file
    try:
        with open(join(WORDS_DIR, WORDS_FOUND_FILE), mode="r", encoding="utf_8") as jsonfile:
            words: Dict[str, Optional[int]] = loads(jsonfile.read())
    except FileNotFoundError:
        # If not found we initialize it with the content of the text file
        words = {word: None for word in get_words_txt() if len(word) == 5}
        save_words(words)

    if ARGS.sort:
        words = dict(OrderedDict(sorted(words.items(), key=lambda item: item[0])))

    # Return the words
    return words


def wordsapi(word: str) -> Optional[int]:
    """
    Get the frequency from the Words API.
    The frequency can be:
    >0 : valid frequency
    0: word known to Words API but no frequency
    None: word unknown to Words API
    """
    url = f"https://wordsapiv1.p.rapidapi.com/words/{word}"

    headers = {
        "x-rapidapi-host": "wordsapiv1.p.rapidapi.com",
        "x-rapidapi-key": getenv("RAPIDAPI_KEY"),
    }

    response = get(url, headers=headers)

    # We got a call
    WORDS_API["counter"] += 1

    if response.status_code == 200:
        frequency: Optional[int] = loads(response.text).get("frequency", 0)
    else:
        frequency = None

    verbose(f"Checked the Words API for '{word}': {frequency}")

    return frequency


def filterapi(selected: Dict[str, Optional[int]], words: Dict[str, Optional[int]]) -> Dict[str, Optional[int]]:
    """
    Check the Words API, but first see if we didn't do it already.
    """

    unknown: List[str] = []

    try:

        # Check each selected words
        for word, frequency in selected.items():

            # We already have the frequency
            if frequency is not None:
                continue

            # We already reached the quota
            if WORDS_API["counter"] >= WORDS_API["quota"]:
                continue

            # Let's ask the API about it.
            frequency = wordsapi(word)

            # Update the dictionaries of words
            if frequency is None:  # Not known by the Words API
                unknown.append(word)  # To be removed later, outside the loop
            else:
                selected[word] = frequency
                words[word] = frequency

        # Remove the words unkown to the API
        for word in unknown:
            words.pop(word)
            selected.pop(word)

        # Update the words file
        save_words(words)

    except KeyboardInterrupt:
        # Remove the words unkown to the API
        for word in unknown:
            words.pop(word)
        # We want to make sure we save the good work
        save_words(words)
        # Tell the user and exit
        print("\nFilter with the Words API stopped manually by you!")
        sys.exit(1)

    return selected


def stats(words: Dict[str, Optional[int]]) -> None:
    """
    Print various stats about the list of words.
    """
    # Count the usage if each letters in the list of words
    counter = {letter: 0 for letter in ascii_letters}
    for word in words.keys():
        for letter in word:
            try:
                counter[letter] += 1
            except KeyError:
                print(f"The word {word} contains non alpha letter: {letter}")

    # Display the 10 most used letters
    countersorted = sorted(counter.items(), key=lambda item: -item[1])
    mostusedletters = [item[0] for item in countersorted[:10]]
    print(f"10 most used letters in the words: {', '.join(mostusedletters)}")

    # Scores each word with the frequency of its letters
    score = {}
    for word in words.keys():
        score[word] = 0
        for letter in word:
            score[word] += counter[letter]

    # Display the 5 most highest scored words
    scoresorted = sorted(score.items(), key=lambda item: -item[1])
    mostscoredword = [item[0] for item in scoresorted[:5]]
    print(f"5 most scored words: {', '.join(mostscoredword)}")


def verbose(text: str) -> None:
    """
    Print a message on the console in verbose mode.
    """
    if ARGS.verbose:
        print(text)


def composedofletters(word: str, letters: str) -> bool:
    """
    Check if a word only contains the provided letters.
    """
    for letter in word:
        if letter not in letters:
            return False
    return True


def filterwords(
    words: Dict[str, Optional[int]], predicate: Callable[[str, Optional[int]], bool]
) -> Dict[str, Optional[int]]:
    """
    Filter a generator of words according to the given predicate.
    """
    return {word: frequency for word, frequency in words.items() if predicate(word, frequency)}


def saveplay(gamefilename: str, game: Game) -> None:
    """
    Records the word that was just played and the result given by Wordle.
    """
    print("Please enter the word played: ")
    word = input()
    if not word:
        print("No play recorded.")
        return
    word = word.lower()
    if len(word) != 5 or not word.isalpha():
        print(f"Invalid word '{word}'. It should be purely alphabetic and 5 characters long.")
        return
    print("Please enter the result:")
    resultstr = input()
    if not resultstr:
        print("No play recorded.")
        return
    try:
        result = [max(min(int(i), 2), 0) for i in resultstr]
    except ValueError:
        print(f"Invalid result '{resultstr}'. It should be a string of 0, 1 and 2 (Ex: 00120).")
        return
    if len(result) != 5:
        print(f"Invalid result '{resultstr}'. There should be 5 values.")
        return

    print(f"Saving result for '{word}': {result}")
    game.played.append(Play(word=word, result=result))
    with open(gamefilename, mode="w", encoding="utf_8") as gamefile:
        dump(GameSchema().dump(game), gamefile, indent=4)

    if result == [2, 2, 2, 2, 2]:
        print("You WON!!!!")


def main() -> None:
    """
    Module's entry point.
    """
    parse_arguments()

    words = get_words()
    selected = words

    if ARGS.letters:
        verbose(f"Only use the letters '{ARGS.letters}'")
        selected = filterwords(selected, lambda word, frequency: composedofletters(word, ARGS.letters))

    if ARGS.unique:
        verbose("Only propose words with no letter repetition")
        selected = filterwords(selected, lambda word, frequency: len(word) == len(set(word)))

    # read the current game status
    game = get_game(GAME_FILE)

    # Are we looking for a valid word or one that uses none of the letters used before?
    if ARGS.none:
        verbose("Keep the words with only new letters")
        selected = filterwords(selected, lambda word, frequency: game.is_none(word))
    else:
        verbose("Keep the valid words only")
        selected = filterwords(selected, lambda word, frequency: game.is_valid(word))

    # Check if words exist with Words API
    if ARGS.check:
        verbose("Check the existence and frequency with the API")
        selected = filterapi(selected, words)

    if ARGS.minfrequency:
        verbose(f"Keep only words of at least {ARGS.minfrequency} frequency")
        selected = filterwords(selected, lambda word, frequency: bool(frequency and frequency >= ARGS.minfrequency))

    # Show the result
    print(f"Here is the list of {len(selected)} possibilities: {selected}")

    # Show the stats
    if ARGS.stats:
        stats(selected)

    # Save the game play if any
    saveplay(GAME_FILE, game)


if __name__ == "__main__":
    main()
